@@ -3,6 +3,8 @@
  *
  * Copyright (C) 2003-2005  Eric Biederman (ebiederm@xmission.com)
  *
+ * Modified (2007-05-15) by Francesco Chiechi to rudely handle mips platform
+ *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
  * the Free Software Foundation (version 2 of the License).
@@ -28,8 +30,15 @@
 #include <sys/stat.h>
 #include <unistd.h>
 #include <fcntl.h>
+#ifndef _O_BINARY
+#define _O_BINARY 0
+#endif
 #include <getopt.h>
-#ifdef HAVE_ZLIB_H
+#include <ctype.h>
+
+#include "config.h"
+
+#ifdef HAVE_LIBZ
 #include <zlib.h>
 #endif
 #include <sha256.h>
@@ -37,9 +46,11 @@
 #include "kexec-syscall.h"
 #include "kexec-elf.h"
 #include "kexec-sha256.h"
+#include <arch/options.h>
 
-static unsigned long long mem_min = 0;
-static unsigned long long mem_max = ULONG_MAX;
+unsigned long long mem_min = 0;
+unsigned long long mem_max = ULONG_MAX;
+unsigned long kexec_flags = 0;
 
 void die(char *fmt, ...)
 {
@@ -57,6 +68,8 @@ void *xmalloc(size_t size)
 {
 	void *buf;
 	buf = malloc(size);
+	if (!size)
+		return NULL;
 	if (!buf) {
 		die("Cannot malloc %ld bytes: %s\n",
 			size + 0UL, strerror(errno));
@@ -75,12 +88,8 @@ void *xrealloc(void *ptr, size_t size)
 	return buf;
 }
 
-
-/* local variables */
-static struct memory_range *memory_range;
-static int memory_ranges;
-
-int valid_memory_range(unsigned long sstart, unsigned long send)
+int valid_memory_range(struct kexec_info *info,
+		       unsigned long sstart, unsigned long send)
 {
 	int i;
 	if (sstart > send) {
@@ -89,13 +98,18 @@ int valid_memory_range(unsigned long sstart, unsigned long send)
 	if ((send > mem_max) || (sstart < mem_min)) {
 		return 0;
 	}
-	for (i = 0; i < memory_ranges; i++) {
+	for (i = 0; i < info->memory_ranges; i++) {
 		unsigned long mstart, mend;
 		/* Only consider memory ranges */
-		if (memory_range[i].type != RANGE_RAM)
+		if (info->memory_range[i].type != RANGE_RAM)
 			continue;
-		mstart = memory_range[i].start;
-		mend = memory_range[i].end;
+		mstart = info->memory_range[i].start;
+		mend = info->memory_range[i].end;
+		if (i < info->memory_ranges - 1
+		    && mend == info->memory_range[i+1].start
+		    && info->memory_range[i+1].type == RANGE_RAM)
+			mend = info->memory_range[i+1].end;
+
 		/* Check to see if we are fully contained */
 		if ((mstart <= sstart) && (mend >= send)) {
 			return 1;
@@ -104,13 +118,14 @@ int valid_memory_range(unsigned long sstart, unsigned long send)
 	return 0;
 }
 
-int valid_memory_segment(struct kexec_segment *segment)
+static int valid_memory_segment(struct kexec_info *info,
+				struct kexec_segment *segment)
 {
 	unsigned long sstart, send;
 	sstart = (unsigned long)segment->mem;
 	send   = sstart + segment->memsz - 1;
 
-	return valid_memory_range(sstart, send);
+	return valid_memory_range(info, sstart, send);
 }
 
 void print_segments(FILE *f, struct kexec_info *info)
@@ -119,10 +134,14 @@ void print_segments(FILE *f, struct kexec_info *info)
 
 	fprintf(f, "nr_segments = %d\n", info->nr_segments);
 	for (i = 0; i < info->nr_segments; i++) {
-		fprintf(f, "segment[%d].buf   = %p\n",	i, info->segment[i].buf);
-		fprintf(f, "segment[%d].bufsz = %zx\n", i, info->segment[i].bufsz);
-		fprintf(f, "segment[%d].mem   = %p\n",	i, info->segment[i].mem);
-		fprintf(f, "segment[%d].memsz = %zx\n", i, info->segment[i].memsz);
+		fprintf(f, "segment[%d].buf   = %p\n",	i,
+			info->segment[i].buf);
+		fprintf(f, "segment[%d].bufsz = %zx\n", i,
+			info->segment[i].bufsz);
+		fprintf(f, "segment[%d].mem   = %p\n",	i,
+			info->segment[i].mem);
+		fprintf(f, "segment[%d].memsz = %zx\n", i,
+			info->segment[i].memsz);
 	}
 }
 
@@ -187,19 +206,20 @@ unsigned long locate_hole(struct kexec_info *info,
 	}
 
 	/* Compute the free memory ranges */
-	max_mem_ranges = memory_ranges + (info->nr_segments -1);
-	mem_range = malloc(max_mem_ranges *sizeof(struct memory_range));
+	max_mem_ranges = info->memory_ranges + info->nr_segments;
+	mem_range = xmalloc(max_mem_ranges *sizeof(struct memory_range));
 	mem_ranges = 0;
 		
 	/* Perform a merge on the 2 sorted lists of memory ranges  */
-	for (j = 0, i = 0; i < memory_ranges; i++) {
-		unsigned long sstart, send;
-		unsigned long mstart, mend;
-		mstart = memory_range[i].start;
-		mend = memory_range[i].end;
-		if (memory_range[i].type != RANGE_RAM)
+	for (j = 0, i = 0; i < info->memory_ranges; i++) {
+		unsigned long long sstart, send;
+		unsigned long long mstart, mend;
+		mstart = info->memory_range[i].start;
+		mend = info->memory_range[i].end;
+		if (info->memory_range[i].type != RANGE_RAM)
 			continue;
-		while ((j < info->nr_segments) && (((unsigned long)info->segment[j].mem) <= mend)) {
+		while ((j < info->nr_segments) &&
+		       (((unsigned long)info->segment[j].mem) <= mend)) {
 			sstart = (unsigned long)info->segment[j].mem;
 			send = sstart + info->segment[j].memsz -1;
 			if (mstart < sstart) {
@@ -211,7 +231,7 @@ unsigned long locate_hole(struct kexec_info *info,
 			mstart = send +1;
 			j++;
 		}
-		if (mstart <= mend) {
+		if (mstart < mend) {
 			mem_range[mem_ranges].start = mstart;
 			mem_range[mem_ranges].end = mend;
 			mem_range[mem_ranges].type = RANGE_RAM;
@@ -232,7 +252,8 @@ unsigned long locate_hole(struct kexec_info *info,
 		if (start < hole_min) {
 			start = hole_min;
 		}
-		start = (start + hole_align - 1) & ~(hole_align - 1);
+		start = (start + hole_align - 1) &
+			~((unsigned long long)hole_align - 1);
 		if (end > mem_max) {
 			end = mem_max;
 		}
@@ -250,26 +271,28 @@ unsigned long locate_hole(struct kexec_info *info,
 				hole_base = start;
 				break;
 			} else {
-				hole_base = (end - hole_size) & ~(hole_align - 1);
+				hole_base = (end - hole_size) &
+					~((unsigned long long)hole_align - 1);
 			}
 		}
 	}
+	free(mem_range);
 	if (hole_base == ULONG_MAX) {
-		fprintf(stderr, "Could not find a free area of memory of %lx bytes...\n",
-			hole_size);
+		fprintf(stderr, "Could not find a free area of memory of "
+			"%lx bytes...\n", hole_size);
 		return ULONG_MAX;
 	}
 	if ((hole_base + hole_size)  > hole_max) {
-		fprintf(stderr, "Could not find a free area of memory below: %lx...\n",
-			hole_max);
+		fprintf(stderr, "Could not find a free area of memory below: "
+			"%lx...\n", hole_max);
 		return ULONG_MAX;
 	}
 	return hole_base;
 }
 
-void add_segment(struct kexec_info *info,
+void add_segment_phys_virt(struct kexec_info *info,
 	const void *buf, size_t bufsz,
-	unsigned long base, size_t memsz)
+	unsigned long base, size_t memsz, int phys)
 {
 	unsigned long last;
 	size_t size;
@@ -296,13 +319,16 @@ void add_segment(struct kexec_info *info,
 	if (base & (pagesize -1)) {
 		die("Base address: %x is not page aligned\n", base);
 	}
-	
+
+	if (phys)
+		base = virt_to_phys(base);
+
 	last = base + memsz -1;
-	if (!valid_memory_range(base, last)) {
+	if (!valid_memory_range(info, base, last)) {
 		die("Invalid memory segment %p - %p\n",
 			(void *)base, (void *)last);
 	}
-	
+
 	size = (info->nr_segments + 1) * sizeof(info->segment[0]);
 	info->segment = xrealloc(info->segment, size);
 	info->segment[info->nr_segments].buf   = buf;
@@ -311,33 +337,133 @@ void add_segment(struct kexec_info *info,
 	info->segment[info->nr_segments].memsz = memsz;
 	info->nr_segments++;
 	if (info->nr_segments > KEXEC_MAX_SEGMENTS) {
-		fprintf(stderr, 
-			"Warning: kernel segment limit reached. This will likely fail\n");
+		fprintf(stderr, "Warning: kernel segment limit reached. "
+			"This will likely fail\n");
 	}
 }
 
-unsigned long add_buffer(struct kexec_info *info,
+unsigned long add_buffer_phys_virt(struct kexec_info *info,
 	const void *buf, unsigned long bufsz, unsigned long memsz,
 	unsigned long buf_align, unsigned long buf_min, unsigned long buf_max,
-	int buf_end)
+	int buf_end, int phys)
 {
 	unsigned long base;
 	int result;
+	int pagesize;
 
 	result = sort_segments(info);
 	if (result < 0) {
 		die("sort_segments failed\n");
 	}
 
+	/* Round memsz up to a multiple of pagesize */
+	pagesize = getpagesize();
+	memsz = (memsz + (pagesize - 1)) & ~(pagesize - 1);
+
 	base = locate_hole(info, memsz, buf_align, buf_min, buf_max, buf_end);
 	if (base == ULONG_MAX) {
 		die("locate_hole failed\n");
 	}
 	
-	add_segment(info, buf, bufsz, base, memsz);
+	add_segment_phys_virt(info, buf, bufsz, base, memsz, phys);
 	return base;
 }
 
+unsigned long add_buffer_virt(struct kexec_info *info, const void *buf,
+			      unsigned long bufsz, unsigned long memsz,
+			      unsigned long buf_align, unsigned long buf_min,
+			      unsigned long buf_max, int buf_end)
+{
+	return add_buffer_phys_virt(info, buf, bufsz, memsz, buf_align,
+				    buf_min, buf_max, buf_end, 0);
+}
+
+static int find_memory_range(struct kexec_info *info,
+			     unsigned long *base, unsigned long *size)
+{
+	int i;
+	unsigned long start, end;
+
+	for (i = 0; i < info->memory_ranges; i++) {
+		if (info->memory_range[i].type != RANGE_RAM)
+			continue;
+		start = info->memory_range[i].start;
+		end = info->memory_range[i].end;
+		if (end > *base && start < *base + *size) {
+			if (start > *base) {
+				*size = *base + *size - start;
+				*base = start;
+			}
+			if (end < *base + *size)
+				*size = end - *base;
+			return 1;
+		}
+	}
+	return 0;
+}
+
+static int find_segment_hole(struct kexec_info *info,
+			     unsigned long *base, unsigned long *size)
+{
+	int i;
+	unsigned long seg_base, seg_size;
+
+	for (i = 0; i < info->nr_segments; i++) {
+		seg_base = (unsigned long)info->segment[i].mem;
+		seg_size = info->segment[i].memsz;
+
+		if (seg_base + seg_size <= *base)
+			continue;
+		else if (seg_base >= *base + *size)
+			break;
+		else if (*base < seg_base) {
+			*size = seg_base - *base;
+			break;
+		} else if (seg_base + seg_size < *base + *size) {
+			*size = *base + *size - (seg_base + seg_size);
+			*base = seg_base + seg_size;
+		} else {
+			*size = 0;
+			break;
+		}
+	}
+	return *size;
+}
+
+int add_backup_segments(struct kexec_info *info, unsigned long backup_base,
+			unsigned long backup_size)
+{
+	unsigned long mem_base, mem_size, bkseg_base, bkseg_size, start, end;
+	unsigned long pagesize;
+
+	pagesize = getpagesize();
+	while (backup_size) {
+		mem_base = backup_base;
+		mem_size = backup_size;
+		if (!find_memory_range(info, &mem_base, &mem_size))
+			break;
+		backup_size = backup_base + backup_size - \
+			(mem_base + mem_size);
+		backup_base = mem_base + mem_size;
+		while (mem_size) {
+			bkseg_base = mem_base;
+			bkseg_size = mem_size;
+			if (sort_segments(info) < 0)
+				return -1;
+			if (!find_segment_hole(info, &bkseg_base, &bkseg_size))
+				break;
+			start = (bkseg_base + pagesize - 1) & ~(pagesize - 1);
+			end = (bkseg_base + bkseg_size) & ~(pagesize - 1);
+			add_segment_phys_virt(info, NULL, 0,
+					      start, end-start, 0);
+			mem_size = mem_base + mem_size - \
+				(bkseg_base + bkseg_size);
+			mem_base = bkseg_base + bkseg_size;
+		}
+	}
+	return 0;
+}
+
 char *slurp_file(const char *filename, off_t *r_size)
 {
 	int fd;
@@ -351,7 +477,7 @@ char *slurp_file(const char *filename, off_t *r_size)
 		*r_size = 0;
 		return 0;
 	}
-	fd = open(filename, O_RDONLY);
+	fd = open(filename, O_RDONLY | _O_BINARY);
 	if (fd < 0) {
 		die("Cannot open `%s': %s\n",
 			filename, strerror(errno));
@@ -370,9 +496,54 @@ char *slurp_file(const char *filename, off_t *r_size)
 		if (result < 0) {
 			if ((errno == EINTR) ||	(errno == EAGAIN))
 				continue;
-			die("read on %s of %ld bytes failed: %s\n",
-				filename, (size - progress)+ 0UL, strerror(errno));
+			die("read on %s of %ld bytes failed: %s\n", filename,
+			    (size - progress)+ 0UL, strerror(errno));
 		}
+		if (result == 0)
+			die("read on %s ended before stat said it should\n", filename);
+		progress += result;
+	}
+	result = close(fd);
+	if (result < 0) {
+		die("Close of %s failed: %s\n", filename, strerror(errno));
+	}
+	return buf;
+}
+
+/* This functions reads either specified number of bytes from the file or
+   lesser if EOF is met. */
+
+char *slurp_file_len(const char *filename, off_t size)
+{
+	int fd;
+	char *buf;
+	off_t progress;
+	ssize_t result;
+
+	if (!filename)
+		return 0;
+	fd = open(filename, O_RDONLY | _O_BINARY);
+	if (fd < 0) {
+		fprintf(stderr, "Cannot open %s: %s\n", filename,
+				strerror(errno));
+		return 0;
+	}
+	buf = xmalloc(size);
+	progress = 0;
+	while(progress < size) {
+		result = read(fd, buf + progress, size - progress);
+		if (result < 0) {
+			if ((errno == EINTR) ||	(errno == EAGAIN))
+				continue;
+			fprintf(stderr, "read on %s of %ld bytes failed: %s\n",
+					filename, (size - progress)+ 0UL,
+					strerror(errno));
+			free(buf);
+			return 0;
+		}
+		if (result == 0)
+			/* EOF */
+			break;
 		progress += result;
 	}
 	result = close(fd);
@@ -383,7 +554,7 @@ char *slurp_file(const char *filename, off_t *r_size)
 	return buf;
 }
 
-#if HAVE_ZLIB_H
+#if HAVE_LIBZ
 char *slurp_decompress_file(const char *filename, off_t *r_size)
 {
 	gzFile fp;
@@ -470,7 +641,8 @@ static void update_purgatory(struct kexec_info *info)
 		if (info->segment[i].mem == (void *)info->rhdr.rel_addr) {
 			continue;
 		}
-		sha256_update(&ctx, info->segment[i].buf, info->segment[i].bufsz);
+		sha256_update(&ctx, info->segment[i].buf,
+			      info->segment[i].bufsz);
 		nullsz = info->segment[i].memsz - info->segment[i].bufsz;
 		while(nullsz) {
 			unsigned long bytes = nullsz;
@@ -480,20 +652,22 @@ static void update_purgatory(struct kexec_info *info)
 			sha256_update(&ctx, null_buf, bytes);
 			nullsz -= bytes;
 		}
-		region[j].start = info->segment[i].mem;
+		region[j].start = (unsigned long) info->segment[i].mem;
 		region[j].len   = info->segment[i].memsz;
 		j++;
 	}
 	sha256_finish(&ctx, digest);
-	elf_rel_set_symbol(&info->rhdr, "sha256_regions", &region, sizeof(region));
-	elf_rel_set_symbol(&info->rhdr, "sha256_digest", &digest, sizeof(digest));
+	elf_rel_set_symbol(&info->rhdr, "sha256_regions", &region,
+			   sizeof(region));
+	elf_rel_set_symbol(&info->rhdr, "sha256_digest", &digest,
+			   sizeof(digest));
 }
 
 /*
  *	Load the new kernel
  */
 static int my_load(const char *type, int fileind, int argc, char **argv,
-	unsigned long kexec_flags)
+		   unsigned long kexec_flags, void *entry)
 {
 	char *kernel;
 	char *kernel_buf;
@@ -501,12 +675,15 @@ static int my_load(const char *type, int fileind, int argc, char **argv,
 	int i = 0;
 	int result;
 	struct kexec_info info;
+	long native_arch;
 	int guess_only = 0;
 
 	memset(&info, 0, sizeof(info));
 	info.segment = NULL;
 	info.nr_segments = 0;
 	info.entry = NULL;
+	info.backup_start = 0;
+	info.kexec_flags = kexec_flags;
 
 	result = 0;
 	if (argc - fileind <= 0) {
@@ -522,7 +699,8 @@ static int my_load(const char *type, int fileind, int argc, char **argv,
 		kernel_buf, kernel_size);
 #endif
 
-	if (get_memory_ranges(&memory_range, &memory_ranges) < 0) {
+	if (get_memory_ranges(&info.memory_range, &info.memory_ranges,
+		info.kexec_flags) < 0) {
 		fprintf(stderr, "Could not get memory layout\n");
 		return -1;
 	}
@@ -559,17 +737,26 @@ static int my_load(const char *type, int fileind, int argc, char **argv,
 			}
 		}
 	}
-	if (file_type[i].load(argc, argv, kernel_buf, kernel_size, &info) < 0) {
+	if (file_type[i].load(argc, argv, kernel_buf,
+			      kernel_size, &info) < 0) {
 		fprintf(stderr, "Cannot load %s\n", kernel);
 		return -1;
 	}
 	/* If we are not in native mode setup an appropriate trampoline */
-	if (arch_compat_trampoline(&info, &kexec_flags) < 0) {
+	native_arch = physical_arch();
+	if (native_arch < 0) {
 		return -1;
 	}
+	info.kexec_flags |= native_arch;
+	if (arch_compat_trampoline(&info) < 0) {
+		return -1;
+	}
+	if (info.kexec_flags & KEXEC_PRESERVE_CONTEXT) {
+		add_backup_segments(&info, mem_min, mem_max - mem_min + 1);
+	}
 	/* Verify all of the segments load to a valid location in memory */
 	for (i = 0; i < info.nr_segments; i++) {
-		if (!valid_memory_segment(info.segment +i)) {
+		if (!valid_memory_segment(&info, info.segment +i)) {
 			fprintf(stderr, "Invalid memory segment %p - %p\n",
 				info.segment[i].mem,
 				((char *)info.segment[i].mem) + 
@@ -583,19 +770,21 @@ static int my_load(const char *type, int fileind, int argc, char **argv,
 	}
 	/* if purgatory is loaded update it */
 	update_purgatory(&info);
+	if (entry)
+		info.entry = entry;
 #if 0
 	fprintf(stderr, "kexec_load: entry = %p flags = %lx\n", 
-		info.entry, kexec_flags);
+		info.entry, info.kexec_flags);
 	print_segments(stderr, &info);
 #endif
 	result = kexec_load(
-		info.entry, info.nr_segments, info.segment, kexec_flags);
+		info.entry, info.nr_segments, info.segment, info.kexec_flags);
 	if (result != 0) {
 		/* The load failed, print some debugging information */
 		fprintf(stderr, "kexec_load failed: %s\n", 
 			strerror(errno));
 		fprintf(stderr, "entry       = %p flags = %lx\n", 
-			info.entry, kexec_flags);
+			info.entry, info.kexec_flags);
 		print_segments(stderr, &info);
 	}
 	return result;
@@ -604,6 +793,14 @@ static int my_load(const char *type, int fileind, int argc, char **argv,
 int k_unload (unsigned long kexec_flags)
 {
 	int result;
+	long native_arch;
+
+	/* set the arch */
+	native_arch = physical_arch();
+	if (native_arch < 0) {
+		return -1;
+	}
+	kexec_flags |= native_arch;
 
 	result = kexec_load(NULL, 0, NULL, kexec_flags);
 	if (result != 0) {
@@ -619,13 +816,12 @@ int k_unload (unsigned long kexec_flags)
  */
 static int my_shutdown(void)
 {
-	char *args[8];
-	int i = 0;
-
-	args[i++] = "shutdown";
-	args[i++] = "-r";
-	args[i++] = "now";
-	args[i++] = NULL;
+	char *args[] = {
+		"shutdown",
+		"-r",
+		"now",
+		NULL
+	};
 
 	execv("/sbin/shutdown", args);
 	execv("/etc/shutdown", args);
@@ -649,9 +845,47 @@ static int my_exec(void)
 	return -1;
 }
 
+static int kexec_loaded(void);
+
+static int load_jump_back_helper_image(unsigned long kexec_flags, void *entry)
+{
+	int result;
+	struct kexec_segment seg;
+
+	memset(&seg, 0, sizeof(seg));
+	result = kexec_load(entry, 1, &seg, kexec_flags);
+	return result;
+}
+
+/*
+ *	Jump back to the original kernel
+ */
+static int my_load_jump_back_helper(unsigned long kexec_flags, void *entry)
+{
+	int result;
+
+	if (kexec_loaded()) {
+		fprintf(stderr, "There is kexec kernel loaded, make sure "
+			"you are in kexeced kernel.\n");
+		return -1;
+	}
+	if (!entry) {
+		fprintf(stderr, "Please specify jump back entry "
+			"in command line\n");
+		return -1;
+	}
+	result = load_jump_back_helper_image(kexec_flags, entry);
+	if (result) {
+		fprintf(stderr, "load jump back kernel failed: %s\n",
+			strerror(errno));
+		return result;
+	}
+	return result;
+}
+
 static void version(void)
 {
-	printf("kexec " VERSION " released " RELEASE_DATE "\n");
+	printf(PACKAGE_STRING " released " PACKAGE_DATE "\n");
 }
 
 void usage(void)
@@ -659,25 +893,35 @@ void usage(void)
 	int i;
 
 	version();
-	printf(
-		"Usage: kexec [OPTION]... [kernel]\n"
-		"Directly reboot into a new kernel\n"
-		"\n"
-		" -h, --help           Print this help.\n"
-		" -v, --version        Print the version of kexec.\n"
-		" -f, --force          Force an immediate kexec, don't call shutdown.\n"
-		" -x, --no-ifdown      Don't bring down network interfaces.\n"
-		"                      (if used, must be last option specified)\n"
-		" -l, --load           Load the new kernel into the current kernel.\n"
-		" -p, --load-panic     Load the new kernel for use on panic.\n"
-		" -u, --unload         Unload the current kexec target kernel.\n"
-		" -e, --exec           Execute a currently loaded kernel.\n"
-		" -t, --type=TYPE      Specify the new kernel is of this type.\n"
-		"     --mem-min=<addr> Specify the lowest memory addres to load code into.\n"
-		"     --mem-max=<addr> Specify the highest memory addres to load code into.\n"
-		"\n"
-		"Supported kernel file types and options: \n"
-		);
+	printf("Usage: kexec [OPTION]... [kernel]\n"
+	       "Directly reboot into a new kernel\n"
+	       "\n"
+	       " -h, --help           Print this help.\n"
+	       " -v, --version        Print the version of kexec.\n"
+	       " -f, --force          Force an immediate kexec,\n"
+	       "                      don't call shutdown.\n"
+	       " -x, --no-ifdown      Don't bring down network interfaces.\n"
+	       "                      (if used, must be last option\n"
+	       "                       specified)\n"
+	       " -l, --load           Load the new kernel into the\n"
+	       "                      current kernel.\n"
+	       " -p, --load-panic     Load the new kernel for use on panic.\n"
+	       " -u, --unload         Unload the current kexec target kernel.\n"
+	       "                      If capture kernel is being unloaded\n"
+	       "                      specify -p with -u.\n"
+	       " -e, --exec           Execute a currently loaded kernel.\n"
+	       " -t, --type=TYPE      Specify the new kernel is of this type.\n"
+	       "     --mem-min=<addr> Specify the lowest memory address to\n"
+	       "                      load code into.\n"
+	       "     --mem-max=<addr> Specify the highest memory address to\n"
+	       "                      load code into.\n"
+	       "     --reuseinitrd    Reuse initrd from first boot.\n"
+	       "     --load-preserve-context Load the new kernel and preserve\n"
+	       "                      context of current kernel during kexec.\n"
+	       "     --load-jump-back-helper Load a helper image to jump back\n"
+	       "                      to original kernel.\n"
+	       "\n"
+	       "Supported kernel file types and options: \n");
 	for (i = 0; i < file_types; i++) {
 		printf("%s\n", file_type[i].name);
 		file_type[i].usage();
@@ -687,28 +931,124 @@ void usage(void)
 	printf("\n");
 }
 
+static int kexec_loaded(void)
+{
+	int ret;
+	FILE *fp;
+
+	fp = fopen("/sys/kernel/kexec_loaded", "r");
+	if (fp == NULL)
+		return -1;
+	fscanf(fp, "%d", &ret);
+	fclose(fp);
+	return ret;
+}
+
+/*
+ * Remove parameter from a kernel command line. Helper function by get_command_line().
+ */
+static void remove_parameter(char *line, const char *param_name)
+{
+	char *start, *end;
+
+	start = strstr(line, param_name);
+
+	/* parameter not found */
+	if (!start)
+		return;
+
+	/*
+	 * check if that's really the start of a parameter and not in
+	 * the middle of the word
+	 */
+	if (start != line && !isspace(*(start-1)))
+		return;
+
+	end = strstr(start, " ");
+	if (!end)
+		*start = 0;
+	else {
+		memmove(start, end+1, strlen(end));
+		*(end + strlen(end)) = 0;
+	}
+}
+
+/*
+ * Returns the contents of the current command line to be used with
+ * --reuse-cmdline option.  The function gets called from architecture specific
+ * code. If we load a panic kernel, that function will strip the
+ * "crashkernel=" option because it does not make sense that the crashkernel
+ * reserves memory for a crashkernel (well, it would not boot since the
+ * amount is exactly the same as the crashkernel has overall memory). Also,
+ * remove the BOOT_IMAGE from lilo (and others) since that doesn't make
+ * sense here any more. The kernel could be different even if we reuse the
+ * commandline.
+ *
+ * The function returns dynamically allocated memory.
+ */
+char *get_command_line(void)
+{
+	FILE *fp;
+	size_t len;
+	char *line = NULL;
+
+	fp = fopen("/proc/cmdline", "r");
+	if (!fp)
+		die("Could not read /proc/cmdline.");
+	getline(&line, &len, fp);
+	fclose(fp);
+
+	if (line) {
+		/* strip newline */
+		*(line + strlen(line) - 1) = 0;
+
+		remove_parameter(line, "BOOT_IMAGE");
+		if (kexec_flags & KEXEC_ON_CRASH)
+			remove_parameter(line, "crashkernel");
+	} else
+		line = strdup("");
+
+	return line;
+}
+
+/* check we retained the initrd */
+void check_reuse_initrd(void)
+{
+	char *line = get_command_line();
+
+	if (strstr(line, "retain_initrd") == NULL)
+		die("unrecoverable error: current boot didn't "
+		    "retain the initrd for reuse.\n");
+
+	free(line);
+}
+
+
 int main(int argc, char *argv[])
 {
 	int do_load = 1;
 	int do_exec = 0;
+	int do_load_jump_back_helper = 0;
 	int do_shutdown = 1;
 	int do_sync = 1;
 	int do_ifdown = 0;
 	int do_unload = 0;
-	unsigned long kexec_flags = 0;
+	int do_reuse_initrd = 0;
+	void *entry = 0;
 	char *type = 0;
 	char *endptr;
 	int opt;
 	int result = 0;
 	int fileind;
 	static const struct option options[] = {
-		KEXEC_OPTIONS
+		KEXEC_ARCH_OPTIONS
 		{ 0, 0, 0, 0},
 	};
 	static const char short_options[] = KEXEC_OPT_STR;
 
 	opterr = 0; /* Don't complain about unrecognized options here */
-	while ((opt = getopt_long(argc, argv, short_options, options, 0)) != -1) {
+	while ((opt = getopt_long(argc, argv, short_options,
+				  options, 0)) != -1) {
 		switch(opt) {
 		case OPT_HELP:
 			usage();
@@ -744,6 +1084,32 @@ int main(int argc, char *argv[])
 			do_ifdown = 1;
 			do_exec = 1;
 			break;
+		case OPT_LOAD_JUMP_BACK_HELPER:
+			do_load = 0;
+			do_shutdown = 0;
+			do_sync = 1;
+			do_ifdown = 1;
+			do_exec = 0;
+			do_load_jump_back_helper = 1;
+			kexec_flags = KEXEC_PRESERVE_CONTEXT;
+			break;
+		case OPT_ENTRY:
+			entry = (void *)strtoul(optarg, &endptr, 0);
+			if (*endptr) {
+				fprintf(stderr,
+					"Bad option value in --load-jump-back-helper=%s\n",
+					optarg);
+				usage();
+				return 1;
+			}
+			break;
+		case OPT_LOAD_PRESERVE_CONTEXT:
+			do_load = 1;
+			do_exec = 0;
+			do_shutdown = 0;
+			do_sync = 1;
+			kexec_flags = KEXEC_PRESERVE_CONTEXT;
+			break;
 		case OPT_TYPE:
 			type = optarg;
 			break;
@@ -757,7 +1123,8 @@ int main(int argc, char *argv[])
 		case OPT_MEM_MIN:
 			mem_min = strtoul(optarg, &endptr, 0);
 			if (*endptr) {
-				fprintf(stderr, "Bad option value in --mem-min=%s\n",
+				fprintf(stderr,
+					"Bad option value in --mem-min=%s\n",
 					optarg);
 				usage();
 				return 1;
@@ -766,17 +1133,35 @@ int main(int argc, char *argv[])
 		case OPT_MEM_MAX:
 			mem_max = strtoul(optarg, &endptr, 0);
 			if (*endptr) {
-				fprintf(stderr, "Bad option value in --mem-max=%s\n",
+				fprintf(stderr,
+					"Bad option value in --mem-max=%s\n",
 					optarg);
 				usage();
 				return 1;
 			}
 			break;
+		case OPT_REUSE_INITRD:
+			do_reuse_initrd = 1;
+			break;
 		default:
 			break;
 		}
 	}
 
+	if ((kexec_flags & KEXEC_ON_CRASH) && !is_crashkernel_mem_reserved()) {
+		printf("Memory for crashkernel is not reserved\n");
+		printf("Please reserve memory by passing ");
+		printf("\"crashkernel=X@Y\" parameter to the kernel\n");
+		die("Then try loading kdump kernel\n");
+	}
+
+	if (do_load && (kexec_flags & KEXEC_PRESERVE_CONTEXT) &&
+	    mem_max == ULONG_MAX) {
+		printf("Please specify memory range used by kexeced kernel\n");
+		printf("to preserve the context of original kernel with \n");
+		die("\"--mem-max\" parameter\n");
+	}
+
 	fileind = optind;
 	/* Reset getopt for the next pass; called in other source modules */
 	opterr = 1;
@@ -784,11 +1169,31 @@ int main(int argc, char *argv[])
 
 	result = arch_process_options(argc, argv);
 
+	/* Check for bogus options */
+	if (!do_load) {
+		while((opt = getopt_long(argc, argv, short_options,
+					 options, 0)) != -1) {
+			if ((opt == '?') || (opt >= OPT_ARCH_MAX)) {
+				usage();
+				return 1;
+			}
+		}
+	}
+
+	if (do_reuse_initrd){
+		check_reuse_initrd();
+		arch_reuse_initrd();
+	}
+
 	if (do_unload) {
 		result = k_unload(kexec_flags);
 	}
 	if (do_load && (result == 0)) {
-		result = my_load(type, fileind, argc, argv, kexec_flags);
+		result = my_load(type, fileind, argc, argv, kexec_flags, entry);
+	}
+	/* Don't shutdown unless there is something to reboot to! */
+	if ((result == 0) && (do_shutdown || do_exec) && !kexec_loaded()) {
+		die("Nothing has been loaded!\n");
 	}
 	if ((result == 0) && do_shutdown) {
 		result = my_shutdown();
@@ -797,12 +1202,14 @@ int main(int argc, char *argv[])
 		sync();
 	}
 	if ((result == 0) && do_ifdown) {
-		extern int ifdown(void);
-		(void)ifdown();
+		ifdown();
 	}
 	if ((result == 0) && do_exec) {
 		result = my_exec();
 	}
+	if ((result == 0) && do_load_jump_back_helper) {
+		result = my_load_jump_back_helper(kexec_flags, entry);
+	}
 
 	fflush(stdout);
 	fflush(stderr);
